/*!
 ************************************************************************
 * \file adaptive_filter.c
 *
 * \brief
 *    adaptive filter routines
 *
 * \author
 *  Yuri Vatis               <vatis@tnt.uni-hannover.de>    \n
 *
 * This software may be used for research purposes only.
 * Universitaet Hannover would also appreciate that all those using the 
 * software or any extension of it in any way comply with the 
 * following condition:
 *   	Including in any technical report, conference or journal paper
 *   	produced which uses the software or any extension of it in
 *    any way, at least one of bibliographic references:
 *    Yuri Vatis, Bernd Edler, Dieu Thanh Nguyen, Jrn Ostermann,
 *		"Motion-and Aliasing-compensated Prediction using a two-dimensional non-separable Adaptive Wiener Interpolation Filter",
 *		Proc. ICIP 2005, IEEE International Conference on Image Processing, Genova, Italy, September 2005.
 *    Yuri Vatis, Bernd Edler, Ingolf Wassermann, Dieu Thanh Nguyen, Jrn Ostermann,
 *		"Coding of Coefficients of two-dimensional non-separable Adaptive Wiener Interpolation Filter", 
 *		Proc. VCIP 2005, SPIE Visual Communication & Image Processing, Beijing, China, July 2005.
 ************************************************************************
 */

 
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "defines.h"
#include "adaptive_filter.h"

#ifdef ADAPTIVE_FILTER

#ifdef EIGHTH_PEL
#include "image.h"
#endif

const static double STANDARD_2D_FILTER[15][36] = {
{   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		1.0/64.0  ,   -5.0/64.0  ,   52.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // a_pos
{   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		1.0/32.0  ,   -5.0/32.0  ,   20.0/32.0  ,   20.0/32.0  ,   -5.0/32.0  ,   1.0/32.0  ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // b_pos
{   0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   52.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    0.0       ,    0.0       ,   0.0
}, // c_pos
{   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // d_pos
{   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // e_pos
{   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
   -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
	 20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // f_pos
{   0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
		1.0/64.0  ,   -5.0/64.0  ,   20.0/64.0  ,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0       ,    0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,   0.0
}, // g_pos
{   0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   20.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   -5.0/32.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    1.0/32.0  ,    0.0       ,    0.0       ,   0.0
}, // h_pos
{   1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
   -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 20.0/2048.0, -100.0/2048.0, 1040.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 -5.0/2048.0,   25.0/2048.0, -260.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 1.0/2048.0,   -5.0/2048.0,   52.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // i_pos
{   1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0,
   -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
	 20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
	 20.0/1024.0, -100.0/1024.0,  400.0/1024.0,  400.0/1024.0, -100.0/1024.0,  20.0/1024.0,
	 -5.0/1024.0,   25.0/1024.0, -100.0/1024.0, -100.0/1024.0,   25.0/1024.0,  -5.0/1024.0,
	 1.0/1024.0,   -5.0/1024.0,   20.0/1024.0,   20.0/1024.0,   -5.0/1024.0,   1.0/1024.0
}, // j_pos
{   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
   -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 20.0/2048.0, -100.0/2048.0,  400.0/2048.0, 1040.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -260.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   52.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // i_pos
{   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   52.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // l_pos
{   0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
    0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,   20.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		1.0/64.0  ,   -5.0/64.0  ,   40.0/64.0  ,   20.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0       ,    0.0       ,   -5.0/64.0  ,    0.0       ,    0.0       ,   0.0       ,
		0.0       ,    0.0       ,    1.0/64.0  ,    0.0       ,    0.0       ,   0.0
}, // m_pos
{   1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0,
   -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 20.0/2048.0, -100.0/2048.0,  400.0/2048.0,  400.0/2048.0, -100.0/2048.0,  20.0/2048.0,
	 52.0/2048.0, -260.0/2048.0, 1040.0/2048.0, 1040.0/2048.0, -260.0/2048.0,  52.0/2048.0,
	 -5.0/2048.0,   25.0/2048.0, -100.0/2048.0, -100.0/2048.0,   25.0/2048.0,  -5.0/2048.0,
	 1.0/2048.0,   -5.0/2048.0,   20.0/2048.0,   20.0/2048.0,   -5.0/2048.0,   1.0/2048.0
}, // n_pos
{   0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0       ,
    0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
		0.0         ,    0.0         ,  0.0     ,   20.0/64.0  ,    0.0       ,   0.0       ,
		1.0/64.0    ,   -5.0/64.0    , 20.0/64.0,   40.0/64.0  ,   -5.0/64.0  ,   1.0/64.0  ,
		0.0         ,    0.0         ,  0.0     ,   -5.0/64.0  ,    0.0       ,   0.0       ,
		0.0         ,    0.0         ,  0.0     ,    1.0/64.0  ,    0.0       ,   0.0
} // o_pos
};

static int TwoDSymmetricPattern[15][SQR_FILTER]  =		// get one filter from another one, if symmetry properties are used (used e.g. in ExtendFilterCoefficients)
{
	{ 0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	12, 13, 14, 15, 16, 17,        // a_pos  - not used actually
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	12, 13, 14, 14, 13, 12,        // b_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	17, 16, 15, 14, 13, 12,        // c_pos
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0  },
	{ 0,  0, 12,  0,  0,  0,
	0,  0, 13,  0,  0,  0,
	0,  0, 14,  0,  0,  0,        // d_pos
	0,  0, 15,  0,  0,  0,
	0,  0, 16,  0,  0,  0,
	0,  0, 17,  0,  0,  0  },
	{ 0,  1,  2,  3,  4,  5,
	1,  7,  8,  9, 10, 11,
	2,  8, 14, 15, 16, 17,        // e_pos
	3,  9, 15, 21, 22, 23,
	4, 10, 16, 22, 28, 29,
	5, 11, 17, 23, 29, 35  },
	{ 0,  1,  2,  2,  1,  0,
	6,  7,  8,  8,  7,  6,
	12, 13, 14, 14, 13, 12,        // f_pos
	18, 19, 20, 20, 19, 18,
	24, 25, 26, 26, 25, 24,
	30, 31, 32, 32, 31, 30  },
	{ 5,  4,  3,  2,  1,  0,
	11, 10,  9,  8,  7,  1,
	17, 16, 15, 14,  8,  2,        // g_pos
	23, 22, 21, 15,  9,  3,
	29, 28, 22, 16, 10,  4,
	35, 29, 23, 17, 11,  5  },
	{ 0,  0, 12,  0,  0,  0,
	0,  0, 13,  0,  0,  0,
	0,  0, 14,  0,  0,  0,        // h_pos
	0,  0, 14,  0,  0,  0,
	0,  0, 13,  0,  0,  0,
	0,  0, 12,  0,  0,  0  },
	{ 0,  6, 12, 18, 24, 30,
	1,  7, 13, 19, 25, 31,
	2,  8, 14, 20, 26, 32,        // i_pos
	2,  8, 14, 20, 26, 32,
	1,  7, 13, 19, 25, 31,
	0,  6, 12, 18, 24, 30  },
	{ 0,  1,  2,  2,  1,  0,
	1,  7,  8,  8,  7,  1,
	2,  8, 14, 14,  8,  2,   			// j_pos
	2,  8, 14, 14,  8,  2,
	1,  7,  8,  8,  7,  1,
	0,  1,  2,  2,  1,  0  },
	{30, 24, 18, 12,  6,  0,
	31, 25, 19, 13,  7,  1,
	32, 26, 20, 14,  8,  2,   			// k_pos
	32, 26, 20, 14,  8,  2,
	31, 25, 19, 13,  7,  1,
	30, 24, 18, 12,  6,  0  },
	{ 0,  0, 17,  0,  0,  0,
	0,  0, 16,  0,  0,  0,
	0,  0, 15,  0,  0,  0,         // l_pos
	0,  0, 14,  0,  0,  0,
	0,  0, 13,  0,  0,  0,
	0,  0, 12,  0,  0,  0  },
	{ 5, 11, 17, 23, 29, 35,
	4, 10, 16, 22, 28, 29,
	3,  9, 15, 21, 22, 23,        // m_pos
	2,  8, 14, 15, 16, 17,
	1,  7,  8,  9, 10, 11,
	0,  1,  2,  3,  4,  5  },
	{30, 31, 32, 32, 31, 30,
	24, 25, 26, 26, 25, 24,
	18, 19, 20, 20, 19, 18,   			// n_pos
	12, 13, 14, 14, 13, 12,
	6,  7,  8,  8,  7,  6,
	0,  1,  2,  2,  1,  0  },
	{35, 29, 23, 17, 11,  5,
	29, 28, 22, 16, 10,  4,
	23, 22, 21, 15,  9,  3,   			// o_pos
	17, 16, 15, 14,  8,  2,
	11, 10,  9,  8,  7,  1,
	5,  4,  3,  2,  1,  0	}
};
int FILTER_NEW_SUB_POS[36] = {       a_pos, b_pos, a_pos,
	a_pos, e_pos, f_pos, e_pos,
	b_pos, f_pos, j_pos, f_pos,
	a_pos, e_pos, f_pos, e_pos };
//static int SymmetryPosition[15] = {1,1,0,0,1,1,0,0,0,1,0,0,0,0};															
	int FILTER_SIZE = 6;
	int FILTER_OFFSET;
	int UseAllSubpelPositions;
	int SubpelPositionsPattern;
	int DiffQFilterCoef[15][SQR_FILTER];						// differences to be received
	int FilterFlag[15];			  											// Flags defining if Filter at the particular position calculated
	int POS_EQUATION_NUMBER[15];										// number of different coefficietns for each sub-pel position
	int NumberOfQBits;															// number of bits used for quantization of
	static double FilterCoef[15][36];			  				// SQR(PEL)-1 X SQR(FILTER_SIZE) - Filter Coeffitients
	static int    QFilterCoef[15][SQR_FILTER];			// SQR(PEL)-1 X SQR(FILTER_SIZE) - quantized filter coefficients
//	static double LastFilterCoef[15][SQR_FILTER];		// filter used for previous frame
	static double PredFilterCoef[15][SQR_FILTER];		// predicted filter coefficients
	static int		QPredFilterCoef[15][SQR_FILTER];	// quantized predicted filter coefficients

/*!
 ************************************************************************
 * \brief
 *    init adaptive filter structure
 ************************************************************************
 */
void InitAdaptiveFilter(void)
{
	FILTER_OFFSET = (FILTER_SIZE)/2-1;

	POS_EQUATION_NUMBER[a_pos] =  6;
	POS_EQUATION_NUMBER[b_pos] =  3;
	POS_EQUATION_NUMBER[c_pos] =  6;
	
	POS_EQUATION_NUMBER[d_pos] =  6;
	POS_EQUATION_NUMBER[e_pos] = 21;
	POS_EQUATION_NUMBER[f_pos] = 18;
	POS_EQUATION_NUMBER[g_pos] = 21;
	
	POS_EQUATION_NUMBER[h_pos] =  3;
	POS_EQUATION_NUMBER[i_pos] = 18;
	POS_EQUATION_NUMBER[j_pos] =  6;
	POS_EQUATION_NUMBER[k_pos] = 18;
	
	POS_EQUATION_NUMBER[l_pos] =  6;
	POS_EQUATION_NUMBER[m_pos] = 21;
	POS_EQUATION_NUMBER[n_pos] = 18;
	POS_EQUATION_NUMBER[o_pos] = 21;
}


/*!
 ************************************************************************
 * \brief
 *    free adaptive filter structure
 ************************************************************************
 */
void FreeAdaptiveFilter(void)
{
	;
}

/*
 ************************************************************************
 * \brief
 *    reset adaptive filter structure
 ************************************************************************
*/
void ResetAdaptiveFilter(void)
{	
	int sub_pel,i;

	for(sub_pel = 0; sub_pel < 15; sub_pel++)
		for(i = 0; i < SQR_FILTER; i++)
		{
			QPredFilterCoef[sub_pel][i] = 0;
			DiffQFilterCoef[sub_pel][i] = 0;
			QFilterCoef    [sub_pel][i] = 0;
			PredFilterCoef [sub_pel][i] = 0.0;
		}
	NumberOfQBits = DEFAULT_QUANT;
}



/*
 ************************************************************************
 * \brief
 *    return if adaptive filter is used for current frame
 ************************************************************************
*/
int UseAdaptiveFilterForCurrentFrame(void)
{
	if(	(img->type == P_SLICE || img->type == B_SLICE) && img->UseAdaptiveFilter == 1)
		return 1;
	else
		return 0;
}

/*!
 ************************************************************************
 * \brief
 *    Predict and calculate filter coefficients
 ************************************************************************
 */
void CalculateFilterCoefficients(void)
{
	int i,j, sub_pel;

	// at first predict b and h position using inter prediction, b and h are already quantized in last frame
	if(img->AdaptiveFilterFlag)
	{
		if(FilterFlag[b_pos])
		{
			for(i=0; i < SQR_FILTER; i++)
			{
				PredFilterCoef[b_pos][i] = STANDARD_2D_FILTER[b_pos][i];
				PredFilterCoef[h_pos][i] = STANDARD_2D_FILTER[h_pos][i];
			}
			DequantizeFilterCoefficients(DiffQFilterCoef[b_pos], FilterCoef[b_pos]);
			// add differences to predicted values
			for(i=FILTER_SIZE*FILTER_OFFSET; i < FILTER_SIZE*(FILTER_OFFSET+1); i++)
			{
				FilterCoef[b_pos][i] += PredFilterCoef[b_pos][i];
			}
		}
		else
		{
			for(i=0; i < SQR_FILTER; i++)
				FilterCoef[b_pos][i] = STANDARD_2D_FILTER[b_pos][i];
		}	
		ExtendFilterCoefficientsFloat(h_pos,FilterCoef);
		
		// predict a_pos:
		if(FilterFlag[a_pos])
		{
			for(i=0; i < SQR_FILTER; i++)
			{
				PredFilterCoef[a_pos][i] = STANDARD_2D_FILTER[a_pos][i];
			}
			DequantizeFilterCoefficients(DiffQFilterCoef[a_pos], FilterCoef[a_pos]);
			// add differences to predicted values
			for(i=FILTER_SIZE*FILTER_OFFSET; i < FILTER_SIZE*(FILTER_OFFSET+1); i++)
			{
				FilterCoef[a_pos][i] += PredFilterCoef[a_pos][i];
			}
		}
		else
		{
			for(i=0; i < SQR_FILTER; i++)
				FilterCoef[a_pos][i] = STANDARD_2D_FILTER[a_pos][i];
		}	
		ExtendFilterCoefficientsFloat(c_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(d_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(l_pos,FilterCoef);
		
		//predict e_pos
		if(FilterFlag[e_pos])
		{
			for(i = 0; i < FILTER_SIZE; i++)
			{
				for(j = 0; j < FILTER_SIZE; j++)
				{
					PredFilterCoef[e_pos][FILTER_SIZE*i+j] =
							FilterCoef[d_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[a_pos][FILTER_SIZE*FILTER_OFFSET+j];
				}
			}
			DequantizeFilterCoefficients(DiffQFilterCoef[e_pos], FilterCoef[e_pos]);
			for(i = 0; i < SQR_FILTER; i++)
			{
				FilterCoef[e_pos][i] += PredFilterCoef[e_pos][i];
			}
		}
		else
		{
			for(i=0; i < SQR_FILTER; i++)
				FilterCoef[e_pos][i] = STANDARD_2D_FILTER[e_pos][i];
		}	
		ExtendFilterCoefficientsFloat(g_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(m_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(o_pos,FilterCoef);
	
		//predict f_pos:
		if(FilterFlag[f_pos])
		{
			for(i = 0; i < FILTER_SIZE; i++)
			{
				for(j = 0; j < FILTER_SIZE; j++)
				{
					PredFilterCoef[f_pos][FILTER_SIZE*i+j] =
							FilterCoef[d_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[b_pos][FILTER_SIZE*FILTER_OFFSET+j];
				}
			}
			DequantizeFilterCoefficients(DiffQFilterCoef[f_pos], FilterCoef[f_pos]);
			for(i = 0; i < SQR_FILTER; i++)
			{
				FilterCoef[f_pos][i] += PredFilterCoef[f_pos][i];
			}
		}
		else
		{
			for(i=0; i < SQR_FILTER; i++)
				FilterCoef[f_pos][i] = STANDARD_2D_FILTER[f_pos][i];
		}	
		ExtendFilterCoefficientsFloat(i_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(k_pos,FilterCoef);
		ExtendFilterCoefficientsFloat(n_pos,FilterCoef);
	
		//predict j_pos:
		if(FilterFlag[j_pos])
		{
			for(i = 0; i < FILTER_SIZE; i++)
			{
				for(j = 0; j < FILTER_SIZE; j++)
				{
					PredFilterCoef[j_pos][FILTER_SIZE*i+j] =
							FilterCoef[h_pos][FILTER_SIZE*i+FILTER_OFFSET]*FilterCoef[b_pos][FILTER_SIZE*FILTER_OFFSET+j];
				}
			}
			DequantizeFilterCoefficients(DiffQFilterCoef[j_pos], FilterCoef[j_pos]);
			for(i = 0; i < SQR_FILTER; i++)
			{
				FilterCoef[j_pos][i] += PredFilterCoef[j_pos][i];
			}
		}
		else
		{
			for(i=0; i < SQR_FILTER; i++)
				FilterCoef[j_pos][i] = STANDARD_2D_FILTER[j_pos][i];
		}	
	}
	else
		for(sub_pel = 0; sub_pel < 15; sub_pel++)
			for(i = 0; i < SQR_FILTER; i++)
				FilterCoef[sub_pel][i] = STANDARD_2D_FILTER[sub_pel][i];
//	PrintFilterCoefInt(DiffQFilterCoef);
//	PrintFilterCoefFloat(PredFilterCoef);
//	PrintFilterCoefFloat(FilterCoef);
}

/*!
 ************************************************************************
 * \brief
 *		inverse quantization of filter coefficients
 *		input:
 *				int QFCoef[][] - filter coefficients to be dequantized
 *		output:
 *				int FCoef[][]   - double filter coefficients
 ************************************************************************
 */
void DequantizeFilterCoefficients(int QFCoef[SQR_FILTER], double FCoef[SQR_FILTER])
{
	double number_of_steps = pow(2,NumberOfQBits-1);  // 1 bit for sign, NumberOfQBits-1 for amplitude
	int j;
	double is = 0.0;
	for(j=0; j < SQR_FILTER; j++)
	{
		is = (double)(QFCoef[j]);
		is = is/(double)number_of_steps;
		FCoef[j]  = is;		// reconstructed (quantized) filter coefficients
	}
}

/*!
 ************************************************************************
 * \brief
 *		quantize filter coefficients
 *		input:
 *				double FCoef[][] - filter coefficients to be quantized
 *		output:
 *				int QFCoef[][]   - quantized filter coefficients, 
 *														 represented in step numbers
 ************************************************************************
 */
void QuantizeFilterCoefficients(double FCoef[SQR_FILTER], int QFCoef[SQR_FILTER])
{
	int number_of_steps = 2^(NumberOfQBits-1);  // 1 bit for sign, NumberOfQBits-1 for amplitude
	int sign;
	int j;
	for(j=0; j < SQR_FILTER; j++)
	{
		sign = FCoef[j] >= 0.0 ? 1:-1;
		QFCoef[j] = sign*(int)(fabs(FCoef[j])*number_of_steps + 0.5);		// step number
		FCoef[j] = (double)(QFCoef[j])/(double)number_of_steps; // reconstructed (quantized) filter coefficients
	}
}


/*!
 ************************************************************************
 * \brief
 *    extend the filter to 2D (FILTER_SIZE X FILTER_SIZE)
 *
 *    input:
 *      sub_pos - sub-position
 *		 input output:
 *			 FCoef[][] - Filter to be extended at symmetric positions,
 *			 							using FILTER_NEW_SUB_POS
 ************************************************************************
 */
void ExtendFilterCoefficientsFloat(int sub_pos, double FCoef[15][SQR_FILTER])
{
	int i;
	for(i = 0; i < SQR_FILTER; i++)
		FCoef[sub_pos][i] = FCoef[FILTER_NEW_SUB_POS[sub_pos]][TwoDSymmetricPattern[sub_pos][i]];
}


/*!
 ************************************************************************
 * \brief
 *    extend the filter to 2D (FILTER_SIZE X FILTER_SIZE)
 *
 *    input:
 *      sub_pos - sub-position
 *		 input output:
 *			 FCoef[][] - Filter to be extended at symmetric positions,
 *			 							using FILTER_NEW_SUB_POS
 ************************************************************************
 */
void ExtendFilterCoefficientsInt(int sub_pos, int FCoef[15][SQR_FILTER])
{
	int i;
	for(i = 0; i < SQR_FILTER; i++)
		FCoef[sub_pos][i] = FCoef[FILTER_NEW_SUB_POS[sub_pos]][TwoDSymmetricPattern[sub_pos][i]];
}

#ifdef EIGHTH_PEL
void GetBlockWith2DAIF_eighth_pel( int             ref_idx,
                                   StorablePicture **list,
                                   int             x_pos,
                                   int             y_pos,
                                   int             mvx,
                                   int             mvy,
                                   int             img_width,
                                   int             img_height,
                                   int             block[BLOCK_SIZE][BLOCK_SIZE])
{
  int block2[BLOCK_SIZE][BLOCK_SIZE];
  int block3[BLOCK_SIZE][BLOCK_SIZE];
  int block4[BLOCK_SIZE][BLOCK_SIZE];
  int dmvx = (mvx%8+8)%8;
  int dmvy = (mvy%8+8)%8;


  if(!(dmvx&1) && !(dmvy&1))                                                              //              1/4-pel
    GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, mvx/2, mvy/2, img_width,img_height, block);
  else                                                                                                                            //              1/$
  {
    if(dmvx&1 && !(dmvy&1))                                                         //              horizontal 1/8 and vertical 1/4
    {
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx+1)/2, mvy/2, img_width,img_height, block);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx-1)/2, mvy/2, img_width,img_height, block2);
      average_block(block, block2);
    }
    else if(!(dmvx&1) && dmvy&1)                                    //              horizontal 1/4 and vertical 1/8
    {
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, mvx/2, (mvy+1)/2, img_width,img_height, block);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, mvx/2, (mvy-1)/2, img_width,img_height, block2);
      average_block(block, block2);
    }
    else if((dmvx+dmvy)%4)// diagonal positions upper left - down right interpolation
    {
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx-1)/2, (mvy-1)/2, img_width,img_height, block);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx+1)/2, (mvy+1)/2, img_width,img_height, block2);

      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx+1)/2, (mvy-1)/2, img_width,img_height, block3);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx-1)/2, (mvy+1)/2, img_width,img_height, block4);
      average_block4(block, block2, block3, block4);
    }
    else // diagonal positions upper right - down left interpolation
    {
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx+1)/2, (mvy-1)/2, img_width,img_height, block);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx-1)/2, (mvy+1)/2, img_width,img_height, block2);

      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx+1)/2, (mvy+1)/2, img_width,img_height, block3);
      GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, (mvx-1)/2, (mvy-1)/2, img_width,img_height, block4);
      average_block4(block, block2, block3, block4);
    }
//    average_block(block, block2);
  }
}

void GetBlockWith2DAIF(int             ref_idx,
                       StorablePicture **list,
                       int             x_pos,
                       int             y_pos,
                       int             mvx,
                       int             mvy,
											 int             img_width,
                       int             img_height,
                       int             block[BLOCK_SIZE][BLOCK_SIZE])
{
  if(img->mv_res)
    GetBlockWith2DAIF_eighth_pel (ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block);
  else
    GetBlockWith2DAIF_quarter_pel(ref_idx, list, x_pos, y_pos, mvx, mvy, img_width,img_height, block);
}

/*!
 ************************************************************************
 * \brief
 *    Interpolation of 1/4 subpixel with 2D adaptive filter
 ************************************************************************
 */
void GetBlockWith2DAIF_quarter_pel(int             ref_idx,
                                   StorablePicture **list,
                                   int             x_pos,
                                   int             y_pos,
                                   int             mvx,
                                   int             mvy,
                                   int             img_width,
                                   int             img_height,
                                   int             block[BLOCK_SIZE][BLOCK_SIZE])
#else
void GetBlockWith2DAIF(int             ref_idx,
                       StorablePicture **list,
                       int             x_pos,
                       int             y_pos,
                       int             mvx,
                       int             mvy,
											 int             img_width,
                       int             img_height,
                       int             block[BLOCK_SIZE][BLOCK_SIZE])
#endif
{
	int i,j, fi, fj;
	int fpos_x, fpos_y;
	double result;
	int is;
  int sub_pos, mvx_sub, mvy_sub;
  if(mvx >= 0)
    mvx_sub = mvx%4;      						// x-sub-coordinate in a 4x4block
  else
    mvx_sub = (4-abs(mvx)%4)%4;
  if(mvy >= 0)
    mvy_sub = mvy%4;      						// y-sub-coordinate in a 4x4block
  else
    mvy_sub = (4-abs(mvy)%4)%4;
  sub_pos = mvx_sub + 4*mvy_sub-1;    // pos 0..14 in a 4x4 block

	for(i = 0; i < BLOCK_SIZE; i++)
		for(j = 0; j < BLOCK_SIZE; j++)
		{
			if(sub_pos != -1)
			{
				result = 0.0;
    		for(fi = 0; fi < FILTER_SIZE; fi++)
    			for(fj = 0; fj < FILTER_SIZE; fj++)
    			{
    				fpos_x = FindPosition(img_width,  i+x_pos, fi,mvx);
    				fpos_y = FindPosition(img_height, j+y_pos, fj,mvy);
    				result += FilterCoef[sub_pos][FILTER_SIZE*fj+fi]*list[ref_idx]->imgY[fpos_y][fpos_x];
    			}
   			is = (int)(0.5+result);
      }
      else
      {
				is = 	list[ref_idx]->imgY[max(0, min(img_height-1,j+y_pos+mvy/4))]
                                 [max(0, min(img_width -1,i+x_pos+mvx/4))];
			}
#ifdef  INTERNAL_BIT_DEPTH_INCREASE
      block[i][j] = max(0, min(((1<<img->bitdepth_luma)-1),is));
#else
			block[i][j] = max(0, min(255,is));
#endif
	  }
}

/*!
 ************************************************************************
 * \brief
 *    finds current position in X-direction
 *    input:
 *      Dim     - X/Y-size of image,
 *      cur_pos - current X/Y-coordinate,
 *      Offset  - offset in X/Y-direction from left filter position (2D-filter)
 *      mv      - mvx/y-motion vector
 ************************************************************************
 */

int FindPosition(int Dim, int cur_pos, int Offset, int mv)
{
  int position;
  if(mv < 0 && abs(mv)%4!=0)
    mv = mv - 4;
  mv = mv/4;
  position = cur_pos - FILTER_OFFSET + Offset + mv;
  return (max(0, min(Dim-1, position)));
}

/*!
 ************************************************************************
 * \brief
 *    decide whether to interpolate with a new or an old filter
 *    input:
 *			mvx - x-motion vector
 *			mvy - y-motion vector
 *    output:
 *			0 - standard filter
 *			1 - adaptive filter
 ************************************************************************
 */
int CountWithNewFilter(int mvx, int mvy)
{
  int sub_pos, mvx_sub, mvy_sub;
  if(mvx >= 0)
    mvx_sub = mvx%4;      						// x-sub-coordinate in a 4x4block
  else
    mvx_sub = (4-abs(mvx)%4)%4;
  if(mvy >= 0)
    mvy_sub = mvy%4;      						// y-sub-coordinate in a 4x4block
  else
    mvy_sub = (4-abs(mvy)%4)%4;
  sub_pos = mvx_sub + 4*mvy_sub-1;    // pos 0..14 in a 4x4 block
  if(sub_pos == -1)
		return 0;
  else if(FilterFlag[sub_pos] == 0)
		return 0;
  else
		return 1;
}


/*!
 ************************************************************************
 * \brief
 *    prints filter coefficient
 ************************************************************************
 */
void PrintFilterCoefFloat(double FCoef[15][SQR_FILTER])
{
  int i,j;
  for(i = a_pos; i <= o_pos; i++)
  {
//    printf("Filter_%c=[\n",97+i);
    printf("FilterFlag[%c]=%d\n",97+i,FilterFlag[i]);
    printf("Filter(%d,1:6,1:6)=[\n",i+1);
    for(j = 0; j < SQR(FILTER_SIZE); j++)
    {
      if(!(j%FILTER_SIZE))
        printf("[");
      printf("%12.8f   ",FCoef[i][j]);
      if(j%FILTER_SIZE == FILTER_SIZE-1)
        printf("];\n");
      else
        printf(", ");
    }
    printf("];\n");
  }
}

/*!
 ************************************************************************
 * \brief
 *    prints filter coefficient
 ************************************************************************
 */
void PrintFilterCoefInt(int FCoef[15][SQR_FILTER])
{
	int i,j;
	for(i = a_pos; i <= o_pos; i++)
	{
		printf("%c_pos:\n",97+i);
		for(j = 0; j < SQR(FILTER_SIZE); j++)
		{
			if(FCoef[i][j] >= 0)
				printf(" ");
			printf("%d   \t",FCoef[i][j]);
			if(j%FILTER_SIZE == 5)
				printf("\n");
		}
		printf("========================================================\n");
	}
}

#endif
